"""Types."""

# fmt: off
from collections.abc import Callable  # noqa: F401
from typing import (  # noqa: F401, UP035
    Any,
    Generic,
    Iterable,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from jax import Array  # noqa: F401

# fmt: on
